#include "tk.h"
#include "aux.h"

#define AUX_MAX_KEY_CNT     200
#define AUX_MAX_MOUSE_CNT   20

static struct {
	int keyField;
	void (*KeyFunc)(void);
} keyTable[AUX_MAX_KEY_CNT];

static struct {
	int mouseField;
	void (*MouseFunc)(AUX_EVENTREC *);
} mouseDownTable[AUX_MAX_MOUSE_CNT], mouseUpTable[AUX_MAX_MOUSE_CNT], mouseLocTable[AUX_MAX_MOUSE_CNT];

static int keyTableCount        = 0;
static int mouseDownTableCount  = 0;
static int mouseUpTableCount    = 0;
static int mouseLocTableCount   = 0;
static GLenum displayModeType   = 0;
static GLenum displayModePolicy = AUX_MINIMUM_CRITERIA;
static int displayModeID        = 0;


static void DefaultHandleReshape(int w, int h)
{
	glViewport(0, 0, w, h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0, (GLdouble)w, 0.0, (GLdouble)h, -1.0, 1.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

static void DefaultHandleExpose(int w, int h)
{
}

static GLenum MouseLoc(int x, int y, GLenum button)
{
	AUX_EVENTREC info;
	GLenum flag;
	int i;

	flag = GL_FALSE;
	
	for(i = 0; i < mouseLocTableCount; i++)
	{
		if((button & AUX_LEFTBUTTON) == mouseLocTable[i].mouseField)
		{
			info.event = AUX_MOUSELOC;
			info.data[AUX_MOUSEX] = x;
			info.data[AUX_MOUSEY] = y;
			info.data[AUX_MOUSESTATUS] = AUX_LEFTBUTTON;
			
			(*mouseLocTable[i].MouseFunc)(&info);
			
			flag |= GL_TRUE;
		}
		
		if((button & AUX_RIGHTBUTTON) == mouseLocTable[i].mouseField)
		{
			info.event = AUX_MOUSELOC;
			info.data[AUX_MOUSEX] = x;
			info.data[AUX_MOUSEY] = y;
			info.data[AUX_MOUSESTATUS] = AUX_RIGHTBUTTON;
			
			(*mouseLocTable[i].MouseFunc)(&info);
			
			flag |= GL_TRUE;
		}
		
		if((button & AUX_MIDDLEBUTTON) == mouseLocTable[i].mouseField)
		{
			info.event = AUX_MOUSELOC;
			info.data[AUX_MOUSEX] = x;
			info.data[AUX_MOUSEY] = y;
			info.data[AUX_MOUSESTATUS] = AUX_MIDDLEBUTTON;
			
			(*mouseLocTable[i].MouseFunc)(&info);
			
			flag |= GL_TRUE;
		}
	}
	
	return flag;
}

static GLenum MouseUp(int x, int y, GLenum button)
{
	AUX_EVENTREC info;
	GLenum flag;
	int i;

	flag = GL_FALSE;
	
	for(i = 0; i < mouseUpTableCount; i++)
	{
		if((button & AUX_LEFTBUTTON) == mouseUpTable[i].mouseField)
		{
			info.event = AUX_MOUSEUP;
			info.data[AUX_MOUSEX] = x;
			info.data[AUX_MOUSEY] = y;
			info.data[AUX_MOUSESTATUS] = AUX_LEFTBUTTON;
			
			(*mouseUpTable[i].MouseFunc)(&info);
			
			flag |= GL_TRUE;
		}
		
		if((button & AUX_RIGHTBUTTON) == mouseUpTable[i].mouseField)
		{
			info.event = AUX_MOUSEUP;
			info.data[AUX_MOUSEX] = x;
			info.data[AUX_MOUSEY] = y;
			info.data[AUX_MOUSESTATUS] = AUX_RIGHTBUTTON;
			
			(*mouseUpTable[i].MouseFunc)(&info);
			
			flag |= GL_TRUE;
		}
		
		if((button & AUX_MIDDLEBUTTON) == mouseUpTable[i].mouseField)
		{
			info.event = AUX_MOUSEUP;
			info.data[AUX_MOUSEX] = x;
			info.data[AUX_MOUSEY] = y;
			info.data[AUX_MOUSESTATUS] = AUX_MIDDLEBUTTON;
			
			(*mouseUpTable[i].MouseFunc)(&info);
			
			flag |= GL_TRUE;
		}
	}
	
	return flag;
}

static GLenum MouseDown(int x, int y, GLenum button)
{
	AUX_EVENTREC info;
	GLenum flag;
	int i;

	flag = GL_FALSE;
	
	for(i = 0; i < mouseDownTableCount; i++)
	{
		if((button & AUX_LEFTBUTTON) == mouseDownTable[i].mouseField)
		{
			info.event = AUX_MOUSEDOWN;
			info.data[AUX_MOUSEX] = x;
			info.data[AUX_MOUSEY] = y;
			info.data[AUX_MOUSESTATUS] = AUX_LEFTBUTTON;
			
			(*mouseDownTable[i].MouseFunc)(&info);
			
			flag |= GL_TRUE;
		}
		
		if((button & AUX_RIGHTBUTTON) == mouseDownTable[i].mouseField)
		{
			info.event = AUX_MOUSEDOWN;
			info.data[AUX_MOUSEX] = x;
			info.data[AUX_MOUSEY] = y;
			info.data[AUX_MOUSESTATUS] = AUX_RIGHTBUTTON;
			
			(*mouseDownTable[i].MouseFunc)(&info);
			
			flag |= GL_TRUE;
		}
		
		if((button & AUX_MIDDLEBUTTON) == mouseDownTable[i].mouseField)
		{
			info.event = AUX_MOUSEDOWN;
			info.data[AUX_MOUSEX] = x;
			info.data[AUX_MOUSEY] = y;
			info.data[AUX_MOUSESTATUS] = AUX_MIDDLEBUTTON;
			
			(*mouseDownTable[i].MouseFunc)(&info);
			
			flag |= GL_TRUE;
		}
	}
	
	return flag;
}

static GLenum KeyDown(int key, GLenum status)
{
	GLenum flag;
	int i;

	flag = GL_FALSE;
	
	if(keyTableCount)
	{
		for(i = 0; i < keyTableCount; i++)
		{
			if(key == keyTable[i].keyField)
			{
				(*keyTable[i].KeyFunc)();
				flag = GL_TRUE;
			}
		}
	}
	
	return flag;
}

void auxExposeFunc(void (*Func)(int, int))
{
	tkExposeFunc(Func);
}

void auxReshapeFunc(void (*Func)(int,int))
{
	tkExposeFunc(Func);
	tkReshapeFunc(Func);
}

void auxIdleFunc(void (*Func)(void))
{
	tkIdleFunc(Func);
}

void auxKeyFunc(int key, void (*Func)(void))
{
	if(keyTableCount >= AUX_MAX_KEY_CNT) return;
	
	keyTable[keyTableCount].keyField = key;
	keyTable[keyTableCount].KeyFunc  = Func;
	
	keyTableCount++;
}

void auxMouseFunc(int mouse, int mode, void (*Func)(AUX_EVENTREC *))
{
	if(mouseLocTableCount >= AUX_MAX_MOUSE_CNT) return;
	
	if(mode == AUX_MOUSEDOWN)
	{
		mouseDownTable[mouseDownTableCount].mouseField = mouse;
		mouseDownTable[mouseDownTableCount].MouseFunc  = Func;
		
		mouseDownTableCount++;
	}
	else if(mode == AUX_MOUSEUP)
	{
		mouseUpTable[mouseUpTableCount].mouseField = mouse;
		mouseUpTable[mouseUpTableCount].MouseFunc  = Func;
		
		mouseUpTableCount++;
	}
	else if(mode == AUX_MOUSELOC)
	{
		mouseLocTable[mouseLocTableCount].mouseField = mouse;
		mouseLocTable[mouseLocTableCount].MouseFunc  = Func;
		
		mouseLocTableCount++;
	} 
}

void auxMainLoop(void (*Func)(void))
{
	tkDisplayFunc(Func);
	tkExec();
}

void auxInitPosition(int x, int y, int width, int height)
{
	tkInitPosition(x, y, width, height);
}

void auxInitDisplayMode(GLenum type)
{
	displayModeType = type;
	tkInitDisplayMode(type);
}

void auxInitDisplayModePolicy(GLenum type)
{
	displayModePolicy = type;
}

GLenum auxInitDisplayModeID(GLint id)
{
	displayModeID = id;

	return 0;
}

GLenum auxInitWindow(char *title)
{
	if(tkInitWindow(title) == GL_FALSE) return GL_FALSE;
	
	tkReshapeFunc(DefaultHandleReshape);
	tkExposeFunc(DefaultHandleExpose);
	tkMouseUpFunc(MouseUp);
	tkMouseDownFunc(MouseDown);
	tkMouseMoveFunc(MouseLoc);
	tkKeyDownFunc(KeyDown);
	
	auxKeyFunc(AUX_ESCAPE, auxQuit);
	
	return GL_TRUE;
}

void auxCloseWindow(void)
{
	tkCloseWindow();
	
	keyTableCount       = 0;
	mouseDownTableCount = 0;
	mouseUpTableCount   = 0;
	mouseLocTableCount  = 0;
}

void auxQuit(void)
{
	tkQuit();
}

void auxSwapBuffers(void)
{
	tkSwapBuffers();
}

GLenum auxGetDisplayModePolicy(void)
{
   return displayModePolicy;
}

GLint auxGetDisplayModeID(void)
{
   return displayModeID;
}

GLenum auxGetDisplayMode(void)
{
   return displayModeType;
}

void auxSetOneColor(int index, float r, float g, float b)
{
	tkSetOneColor(index, r, g, b);
}

void auxSetFogRamp(int density, int startIndex)
{
	tkSetFogRamp(density, startIndex);
}

void auxSetGreyRamp(void)
{
	tkSetGreyRamp();
}

void auxSetRGBMap(int size, float *rgb)
{
	tkSetRGBMap(size, rgb);
}

int auxGetColorMapSize(void)
{
	return tkGetColorMapSize();
}

void auxGetMouseLoc(int *x, int *y)
{
	tkGetMouseLoc(x, y);
}

AGLPixelFormat auxGetPixelFmt(void)
{
	return tkGetPixelFmt();
}
