/*
** Copyright 1994 - 1998, Conix Graphics.
** All Rights Reserved.
** 
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Conix Graphics;
** the contents of this file may not be disclosed to third parties,
** copied or duplicated in any form, in whole or in part, without the
** prior written permission of Conix Graphics.
** 
*/

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <assert.h>
#include <timer.h>

#include "glut.h"
#include "glutint.h"

#ifdef __MWERKS__
	#include "SIOUX.h"
#endif

GLUTtimer *freeTimerList = NULL;

GLUTidleCB __glutIdleFunc = NULL;
GLUTtimer *__glutTimerList = NULL;

#ifdef SUPPORT_FORTRAN
	GLUTtimer *__glutNewTimer;
#endif

GLUTwindow *__glutWindowWorkList = NULL;
unsigned int __glutModifierMask = 0;  /* ~0 implies not in
                                          core input callback. */
void __glutMenuEvent(long	menuAndItem);
void __glutDrag(WindowPtr	window,Point	mouseloc);
void __glutContentClick(WindowPtr window, EventRecord *myEvent);
Boolean IsDAccWindow(WindowPtr window);
void __glutGoAwayBox(WindowPtr window,Point	mouseloc);
void __glutCloseWindow(WindowPtr window);
void __glutZoom(WindowPtr window, Point mouseloc, int part);
void __glutKeyDown(EventRecord *myEvent);
void __glutDiskEvent(EventRecord *myEvent);
void __glutOSEvent(EventRecord *myEvent);
void __glutUpdate(WindowPtr window);
void __glutActivate(WindowPtr window, int myModifiers);
void __glutAboutBox(void);
void __glutNextEvent(void);
void __glutMouseDown(EventRecord *myEvent);
void __glutMouseUp(EventRecord *myEvent);

void glutIdleFunc(GLUTidleCB idleFunc)
{
	__glutIdleFunc = idleFunc;
}

void glutTimerFunc(unsigned int interval, GLUTtimerCB timerFunc, int value)
{
	GLUTtimer *timer, *other;
	GLUTtimer **prevptr;
	double now;

	if(!timerFunc) return;

	if(freeTimerList)
	{
		timer = freeTimerList;
		freeTimerList = timer->next;
	}
	else
	{
		timer = (GLUTtimer *) malloc(sizeof(GLUTtimer));
		if (!timer) __glutFatalError("out of memory.");
	}

	timer->func    = timerFunc;
	timer->timeout = interval;

	timer->value = value;
	timer->next  = NULL;

	now = __glutTime();
	ADD_TIME(timer->timeout, timer->timeout, now);

	prevptr = &__glutTimerList;
	other = *prevptr;

	while (other && IS_AFTER(other->timeout, timer->timeout))
	{
		prevptr = &other->next;
		other = *prevptr;
	}

	timer->next = other;

	#ifdef SUPPORT_FORTRAN
		__glutNewTimer = timer;  /* for Fortran binding! */
	#endif

	*prevptr = timer;
}

static void handleTimeouts(void)
{
	double now;
	GLUTtimer *timer;
	
	if(!__glutCurrentWindow) return;
	
	__glutSetCurrentWindow(__glutCurrentWindow);

	if(__glutTimerList)
	{
		now = __glutTime();
		while(IS_AT_OR_AFTER(__glutTimerList->timeout, now))
		{
			timer = __glutTimerList;
			timer->func(timer->value);
			
			__glutTimerList = timer->next;
			timer->next = freeTimerList;
			
			freeTimerList = timer;
			
			if (!__glutTimerList) break;
		}
	}
}

void __glutPutOnWorkList(GLUTwindow *window, int workMask)
{
	if(workMask)
	{
		if(window->workMask)
		{
			window->workMask |= workMask;
		}
		else
		{
			window->workMask = workMask;
			window->prevWorkWin = __glutWindowWorkList;
			__glutWindowWorkList = window;
		}
	}
}

void __glutRemoveFromWorkList(GLUTwindow *window)
{
	GLUTwindow *win, *prev_win, *next_win;
	
	prev_win = NULL;
	win = __glutWindowWorkList;
	next_win = NULL;
	while(win)
	{
		next_win = win->prevWorkWin;
		
		if(win == window) break;
		
		prev_win = win;
		win = next_win;
	}
		
	if(win)
	{
		if(prev_win)
		{
			prev_win->prevWorkWin = next_win;
		}
		else
		{
			__glutWindowWorkList = next_win;
		}
	}
}

void glutPostRedisplay(void)
{
	if(!__glutCurrentWindow)
	{
		__glutWarning("glutPostRedisplay: no active window");
		return;
	}
	
	__glutPutOnWorkList(__glutCurrentWindow, GLUT_REDISPLAY_WORK);
}

static void __glutDrawGrowIcon(WindowPtr window)
{
	Rect		   iconRect;
	RgnHandle	oldClip;
	GrafPtr		origPort;
		
	GetPort(&origPort);
	SetPort(window);

	if(window)
	{
		oldClip = NewRgn();
	
		SetPort(window);
		GetClip(oldClip);
		
		iconRect = window->portRect;
		iconRect.top = iconRect.bottom - 15;
		iconRect.left = iconRect.right - 15;
		
		ClipRect(&iconRect);

		PenNormal();
		DrawGrowIcon(window);

		SetClip(oldClip);
		DisposeRgn(oldClip);
	}
	
	SetPort(origPort);
}

static void __glutGrow(WindowPtr window, Point mouseloc)
{
	Rect	growBounds;
	long	newSize;
	GrafPtr		origPort;
	GLUTwindow *glut_win;
		
	GetPort(&origPort);
	SetPort(window);
	
	growBounds = (**GetGrayRgn()).rgnBBox;
	growBounds.right -= growBounds.left;
	growBounds.bottom -= growBounds.top;
	growBounds.left = 20;
	growBounds.top = 20;
	
	newSize = GrowWindow(window, mouseloc, &growBounds);
	
	if(newSize)
	{
		SizeWindow(window, newSize & 0x0000FFFF, newSize >> 16, false);
		
		glut_win = __glutGetWindow((CGrafPtr) window);

		__glutWindowReshape(glut_win, LoWord(newSize),HiWord(newSize));
		__glutPutOnWorkList(glut_win, GLUT_REDISPLAY_WORK);
	}
	
	SetPort(origPort);
}

void __glutWindowReshape(GLUTwindow *glut_win, int width,int height)
{
	GrafPtr		origPort;
		
	if(!__glutSetCurrentWindow(glut_win))
		return;

	GetPort(&origPort);
	SetPort((GrafPtr) glut_win->win);
	
	glut_win->width = width;
	glut_win->height = height;
	
	if(glut_win->reshape)
	{
		(*glut_win->reshape)(glut_win->width, glut_win->height);
	}
	
	SetPort(origPort);
}

static void __glutZoom(WindowPtr window, Point mouseloc, int part)
{
	short w, h;
	GrafPtr		origPort;
	GLUTwindow *glut_win;
		
	GetPort(&origPort);
	SetPort(window);
	
	if(TrackBox(window, mouseloc, part))
	{
		ZoomWindow(window, part, false);
		
		w = abs(window->portRect.right - window->portRect.left);
		h = abs(window->portRect.top - window->portRect.bottom);
		
		glut_win = __glutGetWindow((CGrafPtr) window);
		
		__glutWindowReshape(glut_win, w, h);
		__glutPutOnWorkList(glut_win, GLUT_REDISPLAY_WORK);
	}
	
	SetPort(origPort);
}

static void __glutRedraw(GLUTwindow *glut_win)
{
	GrafPtr		origPort;
	
	if(!glut_win->fullscreen)
	{
		GetPort(&origPort);
		SetPort((GrafPtr) glut_win->win);
	}
	
	if(glut_win->display) 
	{
		(*glut_win->display)();
	}
	
	if(!glut_win->fullscreen)
	{
		SetPort(origPort);
	}
}

void __glutMouseDown(EventRecord *event)
{	
	short	    part;
	WindowPtr	window;
	GLUTwindow *glut_win;
		
	part = FindWindow(event->where, &window);
	
	switch(part)
	{
		case inMenuBar:
			__glutMenuEvent(MenuSelect(event->where));
		break;
		case inSysWindow:
			SystemClick(event, window);
		break;
		case inDrag:
			__glutDrag(window, event->where);
		break;
		case inGoAway:
			__glutGoAwayBox(window, event->where);
		break;
		case inGrow:
			__glutGrow(window, event->where);
		break;
		case inZoomIn:
		case inZoomOut:
			__glutZoom(window, event->where, part);
		break;
		case inContent:
			if(!window) return;
			
			glut_win = __glutGetWindow((CGrafPtr) window);
			if(!__glutSetCurrentWindow(glut_win)) return;
	
			if(window != FrontWindow())
			{
				SelectWindow(window);
			}
			else if(glut_win)
			{
				__glutModifierMask = event->modifiers;
				__glutPutOnWorkList(glut_win, GLUT_BUTTON_WORK);
				__glutContentClick(window, event);
			}
		break;
	}
}

void __glutMouseUp(EventRecord	*event)
{
	GrafPtr    origPort;
	int   i;
	
	if(__glutCurrentWindow->fullscreen) return;
	
	GetPort(&origPort);
	
	for(i = 0; i < __glutWindowListSize; i++)
	{
		if(!__glutWindowList[i]) continue;
		
		if(__glutWindowList[i]->mouse && __glutWindowList[i]->button_press)
		{
			SetPort((GrafPtr) __glutWindowList[i]->win);
		    GlobalToLocal(&event->where);
		    
		    if(!__glutSetCurrentWindow(__glutWindowList[i])) continue;
			
			(*__glutWindowList[i]->mouse)(__glutWindowList[i]->button_press - 1, GLUT_UP, event->where.h, event->where.v);
		}
		
		__glutWindowList[i]->button_press = 0;
	}
	
	SetPort(origPort);
}

void __glutDrag(WindowPtr window, Point mouseloc)
{
	GrafPtr		origPort;
	Rect	    dragBounds;
	GLUTwindow *glut_win;
		
	dragBounds = (**GetGrayRgn()).rgnBBox;
	DragWindow(window, mouseloc, &dragBounds);
	
	glut_win = __glutGetWindow((CGrafPtr) window);
	
	if(!__glutSetCurrentWindow(glut_win))
		return;

	/* beretta: added the following to force a redraw when the window is moved */
	GetPort(&origPort);
	SetPort((GrafPtr) glut_win->win);
	
	__glutPutOnWorkList(glut_win, GLUT_REDISPLAY_WORK);
	
	SetPort(origPort);
}

void __glutContentClick(WindowPtr window, EventRecord *myEvent)
{
	GrafPtr		origPort;
	GLUTwindow *glut_win;
	
	if(__glutCurrentWindow->fullscreen)
	{
		if(!__glutCurrentWindow) return;
		
		__glutSetCurrentWindow(__glutCurrentWindow);
		
		return;
	}
	else
	{
		glut_win = __glutGetWindow((CGrafPtr) window);
		
		if(!__glutSetCurrentWindow(glut_win)) return;
	}

	GetPort(&origPort);
	SetPort(window);
	
	GlobalToLocal(&myEvent->where);
	
	__glutModifierMask = myEvent->modifiers;
	__glutModifierMask &= ~shiftKey;    /* shift is not used as a modifier here  */
	__glutModifierMask &= ~optionKey;   /* option is not used as a modifier here  */
	
	if(myEvent->modifiers & shiftKey)
	{
		glut_win->button_press = GLUT_RIGHT_BUTTON + 1;
		if(glut_win->mouse)
		{
			(*glut_win->mouse)(GLUT_RIGHT_BUTTON, GLUT_DOWN, myEvent->where.h, myEvent->where.v);
		}
	}
	else if(myEvent->modifiers & optionKey)
	{
		glut_win->button_press = GLUT_MIDDLE_BUTTON + 1;
		if(glut_win->mouse)
		{
			(*glut_win->mouse)(GLUT_MIDDLE_BUTTON, GLUT_DOWN, myEvent->where.h, myEvent->where.v);
		}
	}
	else
	{
		glut_win->button_press = GLUT_LEFT_BUTTON + 1;
		if(glut_win->mouse)
		{
			(*glut_win->mouse)(GLUT_LEFT_BUTTON, GLUT_DOWN, myEvent->where.h, myEvent->where.v);
		}
	}
	
	SetPort(origPort);
}

Boolean IsDAccWindow(WindowPtr window)
{
	if(window == nil) 
	{ 
		return false;
	}
	else
	{ 
		return(((WindowPeek)window)->windowKind < 0);
	}
}

void __glutGoAwayBox(WindowPtr	window,Point	mouseloc)
{
	GrafPtr		origPort;
		
	GetPort(&origPort);
	SetPort(window);
	
	if(TrackGoAway(window, mouseloc))
	{ 
		__glutCloseWindow(window); 
	}
	
	SetPort(origPort);
}

void __glutCloseWindow(WindowPtr window)
{
	GLUTwindow *glut_win;
	GrafPtr		origPort;
	
	if(__glutCurrentWindow->fullscreen) return;
	
	if(window)
	{
		GetPort(&origPort);
		SetPort(window);
	
		if(IsDAccWindow(window))
		{
			CloseDeskAcc(((WindowPeek)window)->windowKind);
		}
		else 
		{
			glut_win = __glutGetWindow((CGrafPtr) window);
			__glutDestroyWindow(glut_win, glut_win);
		}
		
		SetPort(origPort);
	}
}

void __glutKeyDown(EventRecord	*myEvent)
{ 
	char myCharCode, myKeyCode;
	GrafPtr		origPort;
	GLboolean specasc;
	GLUTwindow  *glut_win;
	
	if(!__glutCurrentWindow) return;
	
	if(__glutCurrentWindow->fullscreen)
	{
		if(!__glutSetCurrentWindow(__glutCurrentWindow)) return;
	}
	else
	{
		glut_win = __glutGetWindow((CGrafPtr) FrontWindow());
		if(!__glutSetCurrentWindow(glut_win)) return;
		
		GetPort(&origPort);
		SetPort((GrafPtr) __glutCurrentWindow->win);
	}
	
	myCharCode	= (char) BitAnd(myEvent->message,charCodeMask);
	myKeyCode	= (char) (BitAnd(myEvent->message, keyCodeMask) >> 8);
	
	if(myEvent->modifiers & cmdKey)
	{
		if (myEvent->what != autoKey)
		{
			__glutMenuKeyCommand();
			__glutMenuEvent(MenuKey(myCharCode));
		}
	}
	else if(__glutCurrentWindow->keyboard || __glutCurrentWindow->special)
	{
	    __glutModifierMask = myEvent->modifiers;
	    
	    GlobalToLocal(&myEvent->where);
		
		if(__glutCurrentWindow->keyboard && (myCharCode >= 0x20 || (myCharCode < 0x1C
			&& myCharCode != 0x01 && myCharCode != 0x04 && myCharCode != 0x05
			&& myCharCode != 0x0B && myCharCode != 0x0C && myCharCode != 0x10)))
		{
			(*__glutCurrentWindow->keyboard)(myCharCode, myEvent->where.h, myEvent->where.v);
		}
		else if(__glutCurrentWindow->special)
		{
			specasc = true;
			switch(myCharCode)
			{
				case 0x1C:
					(*__glutCurrentWindow->special)(GLUT_KEY_LEFT, myEvent->where.h, myEvent->where.v);
				break;
				case 0x1E:
					(*__glutCurrentWindow->special)(GLUT_KEY_UP, myEvent->where.h, myEvent->where.v);
				break;
				case 0x1D:
					(*__glutCurrentWindow->special)(GLUT_KEY_RIGHT, myEvent->where.h, myEvent->where.v);
				break;
				case 0x1F:
					(*__glutCurrentWindow->special)(GLUT_KEY_DOWN, myEvent->where.h, myEvent->where.v);
				break;
				case 0x0B:
					(*__glutCurrentWindow->special)(GLUT_KEY_PAGE_UP, myEvent->where.h, myEvent->where.v);
				break;
				case 0x0C:
					(*__glutCurrentWindow->special)(GLUT_KEY_PAGE_DOWN, myEvent->where.h, myEvent->where.v);
				break;
				case 0x01:
					(*__glutCurrentWindow->special)(GLUT_KEY_HOME, myEvent->where.h, myEvent->where.v);
				break;
				case 0x04:
					(*__glutCurrentWindow->special)(GLUT_KEY_END, myEvent->where.h, myEvent->where.v);
				break;
				case 0x05:
					(*__glutCurrentWindow->special)(GLUT_KEY_INSERT, myEvent->where.h, myEvent->where.v);
				break;
				default:
					specasc = false;
				break;
			}
			
			if(!specasc)
			{
				switch(myKeyCode)
				{
					case 0x7A:
						(*__glutCurrentWindow->special)(GLUT_KEY_F1, myEvent->where.h, myEvent->where.v);
					break;
					case 0x78:
						(*__glutCurrentWindow->special)(GLUT_KEY_F2, myEvent->where.h, myEvent->where.v);
					break;
					case 0x63:
						(*__glutCurrentWindow->special)(GLUT_KEY_F3, myEvent->where.h, myEvent->where.v);
					break;
					case 0x76:
						(*__glutCurrentWindow->special)(GLUT_KEY_F4, myEvent->where.h, myEvent->where.v);
					break;
					case 0x60:
						(*__glutCurrentWindow->special)(GLUT_KEY_F5, myEvent->where.h, myEvent->where.v);
					break;
					case 0x61:
						(*__glutCurrentWindow->special)(GLUT_KEY_F6, myEvent->where.h, myEvent->where.v);
					break;
					case 0x62:
						(*__glutCurrentWindow->special)(GLUT_KEY_F7, myEvent->where.h, myEvent->where.v);
					break;
					case 0x64:
						(*__glutCurrentWindow->special)(GLUT_KEY_F8, myEvent->where.h, myEvent->where.v);
					break;
					case 0x65:
						(*__glutCurrentWindow->special)(GLUT_KEY_F9, myEvent->where.h, myEvent->where.v);
					break;
					case 0x6D:
						(*__glutCurrentWindow->special)(GLUT_KEY_F10, myEvent->where.h, myEvent->where.v);
					break;
					case 0x67:
						(*__glutCurrentWindow->special)(GLUT_KEY_F11, myEvent->where.h, myEvent->where.v);
					break;
					case 0x6F:
						(*__glutCurrentWindow->special)(GLUT_KEY_F12, myEvent->where.h, myEvent->where.v);
					break;
					default:
				 	break;
				}
			}
		}
	}
	
	if(!__glutCurrentWindow->fullscreen) SetPort(origPort);
}

void __glutDiskEvent(EventRecord *myEvent)
{

}

static void	__glutOSEvent(EventRecord	*myEvent)
{

}

static void __glutUpdate(WindowPtr	window)
{ 
	GrafPtr		origPort;
	GLUTwindow *glut_win;
	
	if(!__glutCurrentWindow) return;
	
	if(!__glutCurrentWindow->fullscreen)
	{
		/* This is done to reattach the drawable in case of a color depth change */
		__glutSetCurrentWindow(__glutCurrentWindow);

		GetPort(&origPort);
		SetPort(window);
		
		BeginUpdate(window);
		EndUpdate(window);
		
		glut_win = __glutGetWindow((CGrafPtr) window);
		__glutPutOnWorkList(glut_win, GLUT_REDISPLAY_WORK);
		
		SetPort(origPort);
	}
	else
	{
		__glutPutOnWorkList(__glutCurrentWindow, GLUT_REDISPLAY_WORK);
	}
}

static void __glutActivate(WindowPtr window, int modifiers)
{
	GLUTwindow *glut_win;
		
	if(modifiers & activeFlag != 0)
	{
		if(!__glutCurrentWindow->fullscreen)
		{
			glut_win = __glutGetWindow((CGrafPtr) window);
			__glutSetCurrentWindow(glut_win);
		}
		else
		{
			__glutSetCurrentWindow(__glutCurrentWindow);
		}
		
		__glutRebuildMenuBar();
	}
}

static void __glutAboutBox(void)
{
	DialogPtr	myDialog;
	short		itemHit;

	myDialog = GetNewDialog(kAboutDialog, nil, (WindowPtr) -1);
	ModalDialog(nil, &itemHit);
	DisposeDialog(myDialog);
}

static void __glutMenuEvent(long	menuAndItem)
{
	int			menuNum;
	int			itemNum;
	int			result;
	Str255		DAName;
	Point       point;
	
	DrawMenuBar();
	
	menuNum = HiWord(menuAndItem);
	itemNum = LoWord(menuAndItem);
	
	if(__glutCurrentWindow)
	{
		GLUTwindow  *glut_win;

		if(!__glutCurrentWindow->fullscreen &&
			(glut_win = __glutGetWindow((CGrafPtr) FrontWindow())) != NULL)
		{
			__glutSetCurrentWindow(glut_win);
		}
		else
		{
			__glutSetCurrentWindow(__glutCurrentWindow);
		}
		
		if(__glutCurrentWindow->menu_status)
		{
			GetMouse(&point);
			__glutCurrentWindow->menu_status(GLUT_MENU_IN_USE, point.h, point.v);
		}
		else if(__glutCurrentWindow->menu_state)
		{
			__glutCurrentWindow->menu_state(GLUT_MENU_IN_USE);
		}
		
		if(__glutCurrentWindow->menu_status)
		{
			GetMouse(&point);
			__glutCurrentWindow->menu_status(GLUT_MENU_NOT_IN_USE, point.h, point.v);
		}
		else if(__glutCurrentWindow->menu_state)
		{
			__glutCurrentWindow->menu_state(GLUT_MENU_NOT_IN_USE);
		}
	}
	
	switch(menuNum) 
	{
		case mApple:
		switch(itemNum)
		{
			case iAbout: 
				__glutAboutBox();
			break;
			default: 
				GetMenuItemText(GetMenuHandle(mApple), itemNum, DAName);
				result = OpenDeskAcc(DAName);
			break;
		}
		break;
		case mFile:
			switch(itemNum)
			{
				case iQuit:
					exit(0);
				break;
			}
		break;
		case mEdit:
		break;
		default:
			if(__glutCurrentWindow) __glutMenuCommand(menuNum, itemNum);
		break;
	}

	HiliteMenu(0);
}

static Boolean __glutConsoleEvent(EventRecord *myEvent)
{
	GLboolean flag = GL_FALSE;
	
	#ifdef __MWERKS__
		flag = SIOUXHandleOneEvent(myEvent);
	#endif
	
	return flag;
}

GLboolean __glutPointInWindow(GLUTwindow *window, Point *point)
{
	if(!window->fullscreen)
	{
		GrafPtr win = (GrafPtr) window->win;
		Rect *rect = &win->portRect;
		
		if(win == FrontWindow())
		{
			if(point->v > rect->top && point->v < rect->bottom &&
			   point->h > rect->left && point->h < rect->right)
			{
				return GL_TRUE;
			}
		}
	}
	else
	{
		return GL_FALSE;
	}
	
	return GL_FALSE;
}

static void processEvents(void)
{
	EventRecord myEvent;
	Boolean gotEvent;
	
	gotEvent = WaitNextEvent(everyEvent, &myEvent, 0, nil);
	if(gotEvent && !__glutConsoleEvent(&myEvent))
	{
		switch(myEvent.what)
		{
			case mouseDown:
				__glutMouseDown(&myEvent);
			break;
			case mouseUp:
				__glutMouseUp(&myEvent);
			break;
			case keyDown:
				__glutKeyDown(&myEvent);
			break;
			case keyUp:
				__glutModifierMask = 0;
			break;
			case autoKey:
				__glutKeyDown(&myEvent);
			break;
			case updateEvt:
				__glutUpdate((WindowPtr) myEvent.message);
			break;
			case diskEvt:
				__glutDiskEvent(&myEvent);
			break;
			case activateEvt:
				__glutActivate((WindowPtr)myEvent.message,myEvent.modifiers);
			break;
			case osEvt:
				__glutOSEvent(&myEvent);
			break;
		}
	}
}

static void processWindowWorkList(GLUTwindow *window)
{
	unsigned int workMask;
	Point point;
	GrafPtr		origPort;
	
	/* Recursively call window work */
	if(window->prevWorkWin) processWindowWorkList(window->prevWorkWin);
	window->prevWorkWin = NULL;
	
	/* Get work mask*/
	workMask = window->workMask;
	
	/* Clear work mask */
	window->workMask = GLUT_NO_WORK;
	
	/* Make context current.
	** We do not needlessly set the current window here. */
	if(window != __glutCurrentWindow && !__glutSetCurrentWindow(window)) return;
	
	if(!window->fullscreen)
	{
		GetPort(&origPort);
		SetPort((GrafPtr) window->win);
	}
	
	/* If mouse motion */
	if(window->button_press)
	{
		if(window->motion)
		{
			GetMouse(&point);
			window->motion(point.h, point.v);
		}
		
		__glutPutOnWorkList(window, GLUT_BUTTON_WORK);
	}
	/* Else if passive work */
	else if(window->passive)
	{
		GetMouse(&point);
		window->passive(point.h, point.v);
		
		__glutPutOnWorkList(window, GLUT_PASSIVE_WORK);
	}
	
	/* If entry work */
	if(window->entry)
	{
		GLboolean in_window;
		
		GetMouse(&point);
		
		in_window = __glutPointInWindow(window, &point);
		
		if(window->entry_mode == GLUT_LEFT && in_window)
		{
			window->entry(GLUT_ENTERED);
			window->entry_mode = GLUT_ENTERED;
		}
		else if(window->entry_mode == GLUT_ENTERED && !in_window)
		{
			window->entry(GLUT_LEFT);
			window->entry_mode = GLUT_LEFT;
		}
		
		__glutPutOnWorkList(window, GLUT_ENTRY_WORK);
	}

	/* Do display work */
	if(workMask & GLUT_REDISPLAY_WORK)
	{
		if(window->forceReshape)
		{
			if(window->visibility) window->visibility(GLUT_VISIBLE);
			if(window->reshape)    window->reshape(window->width, window->height);
  
			window->forceReshape = false;
		}
					
		__glutRedraw(window);
	}

	/* Do finish work */
	if(workMask & GLUT_FINISH_WORK)
	{
		glFinish();
	}

	/* Do debug work */ 
	if(workMask & GLUT_DEBUG_WORK)
	{
		GLenum error;

		while((error = glGetError()) != GL_NO_ERROR)
		{
			__glutWarning("GL error: %s", gluErrorString(error));
		}
	}
	
	if(!window->fullscreen) SetPort(origPort);
}

void glutMainLoop(void)
{
	GLUTwindow *glut_win;
	
	if(!__glutWindowListSize)
	{
		__glutFatalUsage("main loop entered with no windows created.");
	}

	for(;;)
	{
		/* Process OS events */
		processEvents();
		
		/* Process work list */
		if(__glutWindowWorkList)
		{
			glut_win = __glutWindowWorkList;
			__glutWindowWorkList = NULL;
			
			processWindowWorkList(glut_win);
		}
		
		/* If idle */
		if(__glutIdleFunc)
		{
			if(__glutCurrentWindow)
			{
				__glutIdleFunc();
			}
		}
		
		/* If timer list */
		if(__glutTimerList) handleTimeouts();
	}
}
